--[[-----------------------------------------------------------------

Title: EnergyMonitor
Initial Creation Date: 2006/12/05
Notes: Based on GonffBar, EnergyWatch_v2 and EnergyWatch_v2A, however completely rewritten.

Author: lokicoyote
File Date: 2008-12-04T23:10:35Z
File Revision: 72
Project Revision: 73
Project Version: v0.10.4

--]]-----------------------------------------------------------------

assert(LibStub)

local kAddon = "EnergyMonitor"
local kSimpleTimer = kAddon .. "_SimpleTimer-1.0"

-- Cache often used global functions
local ipairs, pairs, select, setmetatable, type, unpack = _G.ipairs, _G.pairs, _G.select, _G.setmetatable, _G.type, _G.unpack
local ceil, floor = _G.math.ceil, _G.math.floor
local GetComboPoints	= _G.GetComboPoints
local GetTime		= _G.GetTime
local PlaySound		= _G.PlaySound
local PlaySoundFile	= _G.PlaySoundFile
local UnitAffectingCombat= _G.UnitAffectingCombat
local UnitBuff		= _G.UnitBuff
local UnitClass		= _G.UnitClass
local UnitIsDeadOrGhost	= _G.UnitIsDeadOrGhost
local UnitPower		= _G.UnitPower
local UnitPowerMax	= _G.UnitPowerMax
local UnitPowerType	= _G.UnitPowerType
local SPELL_POWER_ENERGY= _G.SPELL_POWER_ENERGY

EnergyMonitor = LibStub("AceAddon-3.0"):NewAddon(kAddon, "AceConsole-3.0", "AceEvent-3.0")

---------------------------------------------------------------------
-- Library check
---------------------------------------------------------------------

if not LibStub:GetLibrary("AceLocale-3.0", true) then
	EnergyMonitor:Print("AceLocale-3.0 not loaded.  Addon cannot run.")
	return
end
if not LibStub:GetLibrary(kSimpleTimer, true) then
	EnergyMonitor:Print(kSimpleTimer .. " not loaded.  Addon cannot run.")
	return
end

---------------------------------------------------------------------
-- Library constants
---------------------------------------------------------------------

local addon = LibStub("AceAddon-3.0"):GetAddon(kAddon)
local L = LibStub("AceLocale-3.0"):GetLocale(kAddon, true)

---------------------------------------------------------------------
-- Constant values
---------------------------------------------------------------------

local kEnergyTick = 20

local kShowInCombat = 2^0
local kShowOutOfCombat = 2^1
local kShowInStealth = 2^2
local kShowDead = 2^3
local kShowNonEnergy = 2^4
local kShowInVehicle = 2^5

local kShow = {
	  combat = kShowInCombat
	, dead = kShowDead
	, non_energy = kShowNonEnergy
	, out_of_combat = kShowOutOfCombat
	, stealthed = kShowInStealth
	, vehicle = kShowInVehicle
}

-- Description:: Makes a copy of the kShow table which contains all of
--               the kShowXXX flags
-- Expected result: Return a copy of kShow
-- Input: none
-- Output: a copy of the kShow table
function addon:Get_kShow()
	local Show = self.cache.Show
	if not Show then
		Show = {}
		for k, v in pairs(kShow) do
			Show[ k ] = v
		end
		self.cache.Show = Show
	end
	return Show
end

local kShowWhile = {
	  ["never"] = 0
	, ["combat"] = kShowInCombat
	, ["stealthonly"] = kShowInStealth
	, ["stealth"] = kShowInCombat + kShowInStealth
	, ["energy"] = kShowInCombat + kShowInStealth + kShowOutOfCombat
	, ["alive"] = kShowInVehicle + kShowInCombat + kShowInStealth + kShowOutOfCombat + kShowNonEnergy
	, ["always"] = kShowInVehicle + kShowInCombat + kShowInStealth + kShowOutOfCombat + kShowNonEnergy + kShowDead
}

-- Description: Makes a copy of the kShowWhile table which contains all of the
--              setShowWhile values
-- Expected result: Returns a copy of kShowWhile
-- Input: none
-- Output: a copy of the kShowWhile table
function addon:Get_kShowWhile()
	local ShowWhile = self.cache.ShowWhile
	if not ShowWhile then
		ShowWhile = {}
		for k, v in pairs(kShowWhile) do
			ShowWhile[ k ] = v
		end
		self.cache.ShowWhile = ShowWhile
	end
	return ShowWhile
end

local kShown = {}
for k, v in pairs(kShowWhile) do
	kShown[ v ] = k
end

local kFireUseTime = 2^16
local kFireUseEnergy = 2^17
local kFireContinuous = 2^18
local kFireEnergyPerTick	= kFireUseEnergy + 1
local kFirePercentPerTick	= kFireUseEnergy + 2
local kFireTimePerTick		= kFireUseTime + kFireContinuous + 1

local kFire = {
	  [ kFireEnergyPerTick ]	= "EnergyPerTick"
	, [ kFirePercentPerTick ]	= "PercentPerTick"
	, [ kFireTimePerTick ]		= "TimePerTick"
}

-- Description: Makes a copy of the kFire table which contains all of the
--              SetTick values
-- Expected result: Returns a copy of kFire
-- Input: none
-- Output: a copy of the kFire table
function addon:Get_kFire()
	local Fire = self.cache.Fire
	if not Fire then
		Fire = {}
		for k, v in pairs(kFire) do
			Fire[ k ] = v
		end
		self.cache.Fire = Fire
	end
	return Fire
end

---------------------------------------------------------------------
-- Local functions
---------------------------------------------------------------------

-- Description: Tests whether a bit is set in a bit field
-- Expected result: Preform the result and return if the bit is present
-- Input: value composed of a union of bits and a flag of the form 2^n to test
-- Output: true if the flag is present in value
local function isSet(in_value, in_flag)
	local x = floor(in_value / in_flag)
	return floor(x / 2) ~= ceil(x / 2)
end

-- Description: Compute a bit field based on the shown flags
-- Expected result: Combines the flags of in_show to create a when variable
-- Input: flags indicating when to show
-- Output: bit field of all the flags
local function getWhen(in_show)
	local when = 0
	for k, v in pairs(kShow) do
		if in_show[ k ] then
			when = when + v
		end
	end
	return when
end

function getNow(self)
	if kFireUseTime == self.var.fire_using then
		return GetTime()
	elseif kFireUseEnergy == self.var.fire_using then
		return self.var.current.energy
	end
	return 0
end

-- Description: Returns the media filename (and any defaults)
-- Expected result: If LibSharedMedia is available, use that, otherwise just
--                  return the passed value if it is a string.  If using
--                  LibSharedMedia, the lookup numbers using :List or just
--                  :Fetch strings.  Return a global if that is set too.
-- Input: A LibSharedMedia type and either an index into LibSharedMedia or a
--        media key or media filename
-- Output: The global for the media type and requested filename (global may be nil)
local getMediaFilename -- forward declaration
do
	local SharedMedia = LibStub:GetLibrary("LibSharedMedia-3.0", true)

	if SharedMedia then
		function getMediaFilename(in_type, in_key)
			local global, media = SharedMedia:GetGlobal(in_type), nil
			if type(in_key) == "number" then
				local list = SharedMedia:List(in_type)
				if list[ in_key ] then
					media = SharedMedia:Fetch(in_type, list[ in_key ])
				end
			elseif type(in_key) == "string" and "" ~= in_key then
				media = SharedMedia:Fetch(in_type, in_key, true) or in_key
			end
			return global, media
		end
	else
		function getMediaFilename(in_type, in_key)
			return false, type(in_key) == "string" and "" ~= in_key and in_key or nil
		end
	end
end

---------------------------------------------------------------------
-- State functions
---------------------------------------------------------------------

-- Description: Tests whether the player is Alive
-- Expected result: Determine the current non-dead and non-ghost player state
-- Input: none
-- Output: true if the player is alive
local function isAlive()
	return not UnitIsDeadOrGhost("player")
end

-- Description: Tests whether the player is in combat
-- Expected result: Determine the current combat state
-- Input: none
-- Output: true if the player is in combat
local function isInCombat()
	return UnitAffectingCombat("player")
end

-- Description: Tests whether the player is in a vehicle
-- Expected result: Determine the current vehicle state
-- Input: none
-- Output: true if the player is in a vehicle
local function isNotInVehicle()
	return not UnitInVehicle("player")
end

-- Description: Tests whether the player is stealthed
--              Currently (3.0.3), IsStealthed doesn't always return the correct
--              value during UNIT_AURA.  Another alternative includes enumerating
--              the return values of UnitAura("player", i, "HELPFUL|PLAYER|CANCELABLE")
--              and looking for the stealth texture.  A final better method is
--              simply using UnitBuff("player", "Stealth").
-- Expected result: Determines current stealth state
-- Input: none
-- Output: true if the player is stealthed
local function isStealthed()
	return UnitBuff("player", "Stealth")
end

-- Description: Tests whether the player is using energy
-- Expected result: Determine whether the player is using energy or not
-- Input: none
-- Output: true if the player has an energy bar
local function isUsingEnergy()
	return SPELL_POWER_ENERGY == UnitPowerType("player")
end

---------------------------------------------------------------------
-- Sound functions
---------------------------------------------------------------------

local function resetSoundLocation(in_sound)
	in_sound.file = nil
	in_sound.resource = nil
end

local function playSound(in_sound)
	-- in_sound should have the following properties:
	--	file: String, file to play
	--	resource: String, resource to play
	--	enabled: Boolean, whether the should should play (cleared if fail to play anything)
	-- Assume that the sound was enabled initially
	if in_sound.file then
		if PlaySoundFile(in_sound.file) then
			return
		end
	end
	if in_sound.resource then
		addon:Print("Playing " .. in_sound.resource)
		PlaySound(in_sound.resource)
		return
	end
	in_sound.enabled = false
end

---------------------------------------------------------------------
-- Position functions
---------------------------------------------------------------------

local function savePosition(self)
	if not self.db.profile.bar.position then
		self.db.profile.bar.position = { }
	end
	local frame = self.frame.basePosition
	local position = self.db.profile.bar.position
	position.relativeTo, position.x, position.y = select(3, frame:GetPoint())
end

local function updatePosition(self)
	if self:GetFixed() then return end
	
	local frame = self.frame.basePosition
	local position = self.db.profile.bar.position
	frame:ClearAllPoints()
	if not position then
		frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
		savePosition(self)
	else
		frame:SetPoint("CENTER", UIParent, position.relativeTo, position.x, position.y)
	end
end

---------------------------------------------------------------------
-- Update functions
---------------------------------------------------------------------

function isShown(self)
	local state = self.var.state
	local show = self.db.profile.show
	if state.using_energy or show.non_energy then
		if state.alive or show.dead then
			if state.not_in_vehicle or show.vehicle then
				if show.combat and state.in_combat then
					return true
				end
				if show.out_of_combat and not state.in_combat then
					return true
				end
				if show.stealthed and state.stealthed then
					return true
				end
			end
		end
	end
	return false
end

function updateVisibility(self, in_visible)
	if nil == in_visible then
		in_visible = isShown(self)
	end
	if self.frame.base:IsVisible() ~= in_visible then
		if in_visible then
			self.frame.base:Show()
		else
			self.frame.base:Hide()
		end
	end
end

local function updateShown(self)
	self.var.shown = kShown[ getWhen(self.db.profile.show) ] or "custom"
end

local function updateText(self)
	local text = self.db.profile.bar.text
	text = string.gsub(text, "&em", self.var.current.max)
	text = string.gsub(text, "&e", self.var.current.energy)
	text = string.gsub(text, "&c", self.var.current.points)
	self.frame.text:SetText(text)
end

local function updateSpark(self, in_now)
	local timer = self.var.timer
	local delta = (not self.db.profile.bar.invert and timer:GetElapsed(in_now) or timer:GetRemaining(in_now)) / timer:GetLength()
	if not self.var.continuous and 0 >= delta then
		delta = 1
	end
	local bar = self.frame.bar
	bar:SetValue(delta)
	self.frame.spark:SetPoint("CENTER", bar:GetName(), "LEFT", bar:GetWidth() * delta, 0)
end

---------------------------------------------------------------------
-- AceAddon IMPLEMENTED functions
---------------------------------------------------------------------

do
	local function getDefaults(self)
		if not self.var.defaults and self.frame then
			local texture =  self.frame.bar:GetStatusBarTexture()
			self.var.defaults = {
				profile = {
					bar = {
						color = { 1.0, 1.0, 0, 1.0 }
						, invert = false
						, position = nil
						, scale = 1.0
						, text = ENERGY
						, font = self.frame.text:GetFont()
						, width = self.frame.baseSize:GetWidth()
						, height = self.frame.baseSize:GetHeight()
						, edge = self.frame.edge:GetBackdrop().edgeFile
						, border = self.frame.border:GetTexture() or "Interface\\Tooltips\\UI-StatusBar-Border"
						, texture = texture and texture:GetName() or "Interface\\TargetingFrame\\UI-StatusBar"
					}, show = {
						stealthed = true
						, combat = true
						, out_of_combat = true
						, dead = false
						, non_energy = false
						, vehicle = false
					}, sound = {
						enabled = false
						, resource = "SPELLBOOKCLOSE"
					}, timer = {
						fire = kFireTimePerTick
						, energy_per_tick = kEnergyTick
						, percent_per_tick = 1.0
						, time_per_tick = 2.0
					}
				}
			}
		end
		return self.var.defaults
	end

	function addon:OnInitialize()
		self:setupFrames()
	
		self.cache = setmetatable({}, { __mode = "v" })
		self.var = {
			callbacks = {
				dragstart = function() self:OnDragStart() end
				, dragstop = function() self:OnDragStop() end
				, update = function() self:OnUpdate() end
			}
			, continuous = false
			, current = { energy = 0, max = 0, points = 0 }
			, fire_using = nil
			, fixed = false
			, locked = true
			, state = {
				alive = isAlive()
				, in_combat = isInCombat()
				, not_in_vehicle = isNotInVehicle()
				, stealthed = isStealthed()
				, using_energy = isUsingEnergy()
			}
			, timer = LibStub(kSimpleTimer):New()
		}

		-- Setup Database
		self.db = LibStub("AceDB-3.0"):New(kAddon .. "DB")
		self.db:RegisterDefaults(getDefaults(self))

		-- Listen for profile changes
		self.db.RegisterCallback(self, "OnNewProfile", "LoadProfile")
		self.db.RegisterCallback(self, "OnProfileReset", "LoadProfile")
		self.db.RegisterCallback(self, "OnProfileChanged", "LoadProfile")
		self.db.RegisterCallback(self, "OnProfileCopied", "LoadProfile")

		-- Register slash commands
		if self.setupOptions then self:setupOptions() end
	end
end

do
	-- Description: Tests whether the current player can use this addon
	-- Expected result: Checks if the current player uses energy
	-- Input: none
	-- Output: true if useable
	local function canUtilize()
		local _, class = UnitClass("player")
		return "DRUID" == class or "ROGUE" == class
	end

	function addon:OnEnable()
		-- Called when the addon is enabled
		if not canUtilize() then self:Disable() return end

		-- Initialize variables
		local current = self.var.current
		current.max = UnitPowerMax("player", SPELL_POWER_ENERGY)
		current.energy = UnitPower("player", SPELL_POWER_ENERGY)
		current.points = GetComboPoints("player")
		local state = self.var.state
		state.alive = isAlive()
		state.in_combat = isInCombat()
		state.not_in_vehicle = isNotInVehicle()
		state.stealthed = isStealthed()
		state.using_energy = isUsingEnergy()

		-- Initialize frames
		self.frame.base:SetScript("OnUpdate", self.var.callbacks.update)
		self:SetLocked(self:GetLocked())

		local SharedMedia = LibStub:GetLibrary("LibSharedMedia-3.0", true)
		if SharedMedia then
			SharedMedia.RegisterCallback(self, "LibSharedMedia_SetGlobal")
		end

		-- Register events
		-- Events that control text display
		self:RegisterEvent("UNIT_MAXENERGY")
		self:RegisterEvent("UNIT_COMBO_POINTS")
		-- Events that control visibility
		self:RegisterEvent("UNIT_AURA")
		self:RegisterEvent("UNIT_DISPLAYPOWER")
		self:RegisterEvent("UNIT_ENTERED_VEHICLE")
		self:RegisterEvent("UNIT_EXITED_VEHICLE")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_DEAD")
		self:RegisterEvent("PLAYER_ALIVE")
		self:RegisterEvent("PLAYER_UNGHOST")
		
		self:LoadProfile()
	end
end

function addon:OnDisable()
	self.frame.base:SetScript("OnUpdate", nil)

	local SharedMedia = LibStub:GetLibrary("LibSharedMedia-3.0", true)
	if SharedMedia then
		SharedMedia.UnregisterAllCallbacks(self)
	end
end

---------------------------------------------------------------------
-- AceDB Handlers
---------------------------------------------------------------------

function addon:LoadProfile()
	-- Set the timer up first
	self:setTickSafe(self:GetTick())
	-- Set these after the tick has been loaded
	self:setEnergyPerTickSafe(self:GetEnergyPerTick())
	self:setPercentPerTickSafe(self:GetPercentPerTick())
	self:setTimePerTickSafe(self:GetTimePerTick())

	-- Set frame options
	self:setBarFontSafe(self:GetBarFont())
	self:setBarColorSafe(self:GetBarColor())
	self:SetBarInverted(self:GetBarInverted())
	self:SetBarText(self:GetBarText())
	
	-- Set frame textures	
	self:setBarBorderSafe(self:GetBarBorder())
	-- Bar Edge overrides Bar Border
	self:setBarEdgeSafe(self:GetBarEdge())
	self:setBarTextureSafe(self:GetBarTexture())

	-- Set frame position
	self:setBarScaleSafe(self:GetBarScale())
	self:setBarWidthSafe(self:GetBarWidth())
	self:setBarHeightSafe(self:GetBarHeight())
	updatePosition(self)

	updateShown(self)
	updateVisibility(self)
end

---------------------------------------------------------------------
-- LibSharedMedia Handlers
---------------------------------------------------------------------

function addon:LibSharedMedia_SetGlobal(_, in_media, in_key)
	if "border" == in_media then
		self:setBarTextureSafe(self:GetBarTexture())
	elseif "font" == in_media then
		self:setBarFontSafe(self:GetBarFont())
	elseif "statusbar" == in_media then
		self:setBarEdgeSafe(self:GetBarEdge())
	end
end

---------------------------------------------------------------------
-- Event Handlers
---------------------------------------------------------------------

-- Description: Event fired during frame update
-- Expected result: Update the current energy, the timer countdown, the bar text
--                  and spark position.  If the timer expires, play a sound if
--                  sounds are enabled.
-- Input: none
-- Output: none
function addon:OnUpdate()
	self.var.current.energy = UnitPower("player", SPELL_POWER_ENERGY)

	local now = getNow(self)

	if self.var.timer:IsExpired(now) then
		-- Reset countdown
		self.var.timer:Reset(now)
		if self.db.profile.sound.enabled then
			playSound(self.db.profile.sound)
		end
	elseif not self.var.timer:IsStarted(now) then
		self.var.timer:Reset(now)
	end

	updateText(self)
	updateSpark(self, now)
end

function addon:OnDragStart()
	if not self.var.locked and not self:GetFixed() then
		self.frame.basePosition:StartMoving()
	end
end
			
function addon:OnDragStop()
	if not self:GetFixed() then
		self.frame.basePosition:StopMovingOrSizing()
		savePosition(self)
	end
end

-- Description: Event fired when max energy changes
-- Expected result: Update cached maximum energy, percent per tick values and
--                  bar text
-- Input: The target of the event
-- Output: none
function addon:UNIT_MAXENERGY(_, in_target)
	if in_target ~= "player" then return end

	self.var.current.max = UnitPowerMax("player", SPELL_POWER_ENERGY)
	self:SetPercentPerTick(self:GetPercentPerTick())
	updateText(self)
end

-- Description: Event fired when combo points change
-- Expected result: Update cached combo points and bar text
-- Input: The target of the event
-- Output: none
function addon:UNIT_COMBO_POINTS(_, in_target)
	if in_target ~= "player" then return end
	
	self.var.current.points = GetComboPoints("player")
	updateText(self)
end

-- Description: Event fired when auras change
-- Expected result: Update stealth state and visibility
-- Input: The target of the event
-- Output: none
function addon:UNIT_AURA(_, in_target)
	if in_target ~= "player" then return end

	local stealthed = isStealthed()
	if self.var.state.stealthed ~= stealthed then
		self.var.state.stealthed = stealthed
		updateVisibility(self)
	end
end

-- Description: Event fired when the "mana" type changes
-- Expected result: Update visibility
-- Input: The target of the event
-- Output: none
function addon:UNIT_DISPLAYPOWER(_, in_target)
	if in_target ~= "player" then return end

	local using_energy = isUsingEnergy()
	if self.var.state.using_energy ~= using_energy then
		self.var.state.using_energy = using_energy
		updateVisibility(self)
	end
end

function addon:UNIT_ENTERED_VEHICLE(_, in_target)
	--[===[@alpha@
	self:Print("UNIT_ENTERED_VEHICLE: " .. in_target)
	--@end-alpha@]===]
	if in_target ~= "player" then return end

	self.var.state.not_in_vehicle = false -- isNotInVehicle() returns false
	updateVisibility(self)
end

function addon:UNIT_EXITED_VEHICLE(_, in_target)
	--[===[@alpha@
	self:Print("UNIT_EXITED_VEHICLE: " .. in_target)
	--@end-alpha@]===]
	if in_target ~= "player" then return end

	self.var.state.not_in_vehicle = true -- isNotInVehicle() returns true
	updateVisibility(self)
end

-- Description: Event fired after ending combat
-- Expected result: Update visibility
-- Input: none
-- Output: none
function addon:PLAYER_REGEN_ENABLED()
	self.var.state.in_combat = false -- isInCombat() returns false
	updateVisibility(self)
end

-- Description: Event fired after ending combat
-- Expected result: Update visibility
-- Input: none
-- Output: none
function addon:PLAYER_REGEN_DISABLED()
	self.var.state.in_combat = true -- isInCombat() returns true
	updateVisibility(self)
end

-- Description: Event fired when the player has died
-- Expected result: Hides the bar
-- Input: none
-- Output: none
function addon:PLAYER_DEAD()
	self.var.state.alive = false -- isAlive() returns false
	updateVisibility(self)
end

-- Description: Event fired when the player is alive without being a ghost
-- Expected result: Update visibility
-- Input: none
-- Output: none
function addon:PLAYER_ALIVE()
	local alive = isAlive()
	if self.var.state.alive ~= alive then
		self.var.state.alive = alive
		updateVisibility(self)
	end
end

-- Description: Event fired when the player is alive after *being* a ghost
-- Expected result: Update visibility
-- Input: none
-- Output: none
function addon:PLAYER_UNGHOST()
	self.var.state.alive = true -- isAlive() returns true
	updateVisibility(self)
end

---------------------------------------------------------------------
-- Getter and Setter functions
---------------------------------------------------------------------

function addon:SetBarColor(...)
	assert(self:setBarColorSafe(...))
end

function addon:setBarColorSafe(...)
	for i = 1, (3 >= select('#', ...) and 3 or 4)  do
		local color = select(i, ...)
		if not (0 <= color and color <= 1) then
			return false, L["Error: Invalid color value, was X"](color)
		end
	end
	self.db.profile.bar.color = { ... }
	local r, g, b, a = ...
	if self.frame.bar then
		self.frame.bar:SetStatusBarColor(r, g, b)
	end
	if self.frame.base and nil ~= a then
		self.frame.base:SetAlpha(a)
	end
	return true
end

function addon:GetBarColor()
	return unpack(self.db.profile.bar.color)
end

function addon:SetBarInverted(in_invert)
	self.db.profile.bar.invert = in_invert
	updateSpark(self, getNow(self))
end

function addon:GetBarInverted()
	return self.db.profile.bar.invert
end

function addon:SetBarText(in_text)
	self.db.profile.bar.text = in_text
	updateText(self)
end

function addon:GetBarText()
	return self.db.profile.bar.text
end

function addon:SetBarFont(in_font)
	assert(self:setBarFontSafe(in_font))
end

function addon:setBarFontSafe(in_font)
	local override, media = getMediaFilename("font", in_font)
	local font = override or media
	local _, height, flags = self.frame.text:GetFont()
	if not (font and self.frame.text:SetFont(font, floor(height + 0.5), flags)) then
		return false, L["Error: Unable to load font X"](font or in_font)
	end
	self.db.profile.bar.font = media
	return true
end

function addon:GetBarFont()
	return self.db.profile.bar.font
end

do
----------------------------------
-- Texture functions
----------------------------------

	-- Description: Tests if a texture filename is a valid texture
	-- Expected result: Load the texture to a temporary texture and test if the
	--                  loaded file matches the requested filename or if the
	--                  special value "Interface\\None" was used
	-- Input: A texture filename
	-- Output: true if the texture can be loaded or if the filename is
	--         "Interface\\None"
	local canLoadTexture -- forward declaration
	do 
		local loader = UIParent:CreateTexture()
		loader:Hide()

		canLoadTexture = function(in_texture)
			loader:SetTexture(in_texture)
			local loaded = in_texture and string.find(in_texture, (loader:GetTexture() or ("X" .. in_texture)), 0, true) or in_texture == "Interface\\None"
			loader:SetTexture(nil)
			return loaded
		end
	end

	function addon:SetBarBorder(in_texture)
		assert(self:setBarBorderSafe(in_texture))
	end

	function addon:setBarBorderSafe(in_texture)
		-- We're looking for an actual texture, whereas "border"s are actually backdrop edges
		local texture = type(in_texture) == "string" and "" ~= in_texture and in_texture
		-- only try using in_texture if it is a non-empty string
		if texture and not canLoadTexture(texture) then
			return false, L["Error: Unable to load texture X"](texture or in_texture)
		end

		self.db.profile.bar.border = texture
		local frame = self.frame.border
		if frame then
			if not texture then
				frame:Hide()
				self:SetBarEdge(self:GetBarEdge())
			else
				self.frame.edge:Hide()
				frame:SetTexture(texture)
				frame:Show()
			end
		end
		return true
	end

	function addon:GetBarBorder()
		return self.db.profile.bar.border
	end

	function addon:SetBarEdge(in_texture)
		assert(self:setBarEdgeSafe(in_texture))
	end

	function addon:setBarEdgeSafe(in_texture)
		local override, media = getMediaFilename("border", in_texture)
		local texture = override or media
		-- only try using in_texture if getMediaFilename returned a meaningful value
		if texture and not canLoadTexture(texture) then
			return false, L["Error: Unable to load texture X"](texture or in_texture)
		end

		self.db.profile.bar.edge = media
		local frame = self.frame.edge
		if frame then
			if not texture then
				frame:Hide()
				self:SetBarBorder(self:GetBarBorder())
			else
				self.frame.border:Hide()
				local backdrop = frame:GetBackdrop()
				backdrop.edgeFile = texture
				frame:SetBackdrop(backdrop)
				frame:Show()
			end
		end
		return true
	end

	function addon:GetBarEdge()
		return self.db.profile.bar.edge
	end

	function addon:SetBarTexture(in_texture)
		assert(self:setBarTextureSafe(in_texture))
	end

	function addon:setBarTextureSafe(in_texture)
		local override, media = getMediaFilename("statusbar", in_texture)
		local texture = override or media
		if not canLoadTexture(texture) then
			return false, L["Error: Unable to load texture X"](texture or in_texture)
		end
		self.db.profile.bar.texture = texture
		local frame = self.frame.bar
		if frame then
			frame:SetStatusBarTexture(texture)
		end
		return true
	end

	function addon:GetBarTexture()
		return self.db.profile.bar.texture
	end
end

----------------------------------
-- Size and Position functions
----------------------------------

function addon:SetFixed(in_fixed)
	self.var.fixed = in_fixed
end

function addon:GetFixed()
	return self.var.fixed
end

function addon:SetLocked(in_locked)
	self.var.locked = in_locked
	local frame = self.frame.basePosition
	frame:SetMovable(not in_locked)
	frame:EnableMouse(not in_locked)
	if in_locked then
		frame:RegisterForDrag()
		frame:SetScript("OnDragStart", nil)
		frame:SetScript("OnDragStop", nil)
	else
		local callbacks = self.var.callbacks
		frame:RegisterForDrag("LeftButton")
		frame:SetScript("OnDragStart", callbacks.dragstart)
		frame:SetScript("OnDragStop", callbacks.dragstop)
	end
end

function addon:GetLocked()
	return self.var.locked
end

function addon:SetBarScale(in_scale)
	assert(self:setBarScaleSafe(in_scale))
end

function addon:setBarScaleSafe(in_scale)
	if not self:GetFixed() then
		if not (0 < (in_scale or 0)) then
			return false, L["Error: Bar scale must be a positive value, was X"](in_scale)
		end
		self.db.profile.bar.scale = in_scale
		if self.frame.baseScale then
			self.frame.baseScale:SetScale(in_scale)
		end
	end
	return true
end

function addon:GetBarScale()
	return self.db.profile.bar.scale
end

function addon:SetBarWidth(in_width)
	assert(self:setBarWidthSafe(in_width))
end

function addon:setBarWidthSafe(in_width)
	if not self:GetFixed() then
		if not (50 <= (in_width or 0)) then
			return false, L["Error: Bar width must be at least X, was Y"](50, in_width)
		end
		self.db.profile.bar.width = in_width
		if self.frame.baseSize then
			self.frame.baseSize:SetWidth(in_width)
		end
	end
	return true
end

function addon:GetBarWidth()
	return self.db.profile.bar.width
end

function addon:SetBarHeight(in_height)
	assert(self:setBarHeightSafe(in_height))
end

function addon:setBarHeightSafe(in_height)
	if not self:GetFixed() then
		if not (10 <= (in_height or 0)) then
			return false, L["Error: Bar height must be at least X, was Y"](10, in_height)
		end
		self.db.profile.bar.height = in_height
		if self.frame.baseSize then
			self.frame.baseSize:SetHeight(in_height)
		end
	end
	return true
end

function addon:GetBarHeight()
	return self.db.profile.bar.height
end

----------------------------------
-- Bar update functions
----------------------------------

function addon:SetTick(in_fire)
	assert(self:setTickSafe(in_fire))
end

function addon:setTickSafe(in_fire)
	if nil == kFire[ in_fire ] then
		return false, L["Error: Unknown fire tick value X"](in_fire)
	end
	self.db.profile.timer.fire = in_fire
	if isSet(in_fire, kFireUseTime) then
		self.var.fire_using = kFireUseTime
		self.var.timer:Start(GetTime())
	elseif isSet(in_fire, kFireUseEnergy) then
		self.var.fire_using = kFireUseEnergy
		self.var.timer:Start(0)
	end
	self.var.continuous = isSet(in_fire, kFireContinuous)
	return true
end

function addon:GetTick()
	return self.db.profile.timer.fire
end

function addon:SetEnergyPerTick(in_energy)
	assert(self:setEnergyPerTickSafe(in_energy))
end

function addon:setEnergyPerTickSafe(in_energy)
	if not (0 < (in_energy or 0)) then
		return false, L["Error: Energy per tick must be a positive value, was X"](in_energy)
	end
	self.db.profile.timer.energy_per_tick = in_energy
	if self:GetTick() == kFireEnergyPerTick then
		self.var.timer:SetLength(in_energy)
	end
	return true
end

function addon:GetEnergyPerTick()
	return self.db.profile.timer.energy_per_tick
end

function addon:SetPercentPerTick(in_percent)
	assert(self:setPercentPerTickSafe(in_percent))
end

function addon:setPercentPerTickSafe(in_percent)
	if not (0 < (in_percent or 0) and in_percent <= 1) then
		return false, L["Error: Energy percent per tick must be a positive value (0, 1], was X"](in_percent)
	end
	self.db.profile.timer.percent_per_tick = in_percent
	if self:GetTick() == kFirePercentPerTick then
		self.var.timer:SetLength(self.var.current.max * in_percent)
	end
	return true
end

function addon:GetPercentPerTick()
	return self.db.profile.timer.percent_per_tick
end

function addon:SetTimePerTick(in_time)
	assert(self:setTimePerTickSafe(in_time))
end

function addon:setTimePerTickSafe(in_time)
	if not (0.1 <= (in_time or 0)) then
		return false, L["Error: Time per tick must be a positive value at least 0.1, was X"](in_time)
	end
	self.db.profile.timer.time_per_tick = in_time
	if self:GetTick() == kFireTimePerTick then
		self.var.timer:SetLength(in_time)
	end
	return true
end

function addon:GetTimePerTick()
	return self.db.profile.timer.time_per_tick
end

----------------------------------
-- Visibility functions
----------------------------------

function addon:SetShownWhile(in_when)
	local show = self.db.profile.show
	local showing = kShowWhile[ in_when ] or getWhen(show)
	for k, v in pairs(kShow) do
		show[ k ] = isSet(showing, v)
	end
	updateShown(self)
	updateVisibility(self)
end

function addon:GetShownWhile()
	return self.var.shown
end

function addon:SetShowWhen(in_when, in_enable)
	self.db.profile.show[ in_when ] = in_enable
	updateShown(self)
	updateVisibility(self)
end

function addon:GetShowWhen(in_when)
	return self.db.profile.show[ in_when ]
end

----------------------------------
-- Sound functions
----------------------------------

function addon:SetSoundEnabled(in_enabled)
	self.db.profile.sound.enabled = in_enabled
end

function addon:GetSoundEnabled()
	return self.db.profile.sound.enabled
end

function addon:SetSoundFile(in_file)
	resetSoundLocation(self.db.profile.sound)
	-- We don't care about sound overrides
	self.db.profile.sound.file = select(2, getMediaFilename("sound", in_file))
end

function addon:GetSoundFile()
	return self.db.profile.sound.file
end

function addon:SetSoundResource(in_resource)
	resetSoundLocation(self.db.profile.sound)
	-- We don't care about sound overrides
	self.db.profile.sound.resource = select(2, getMediaFilename("playsound", in_resource))
end

function addon:GetSoundResource()
	return self.db.profile.sound.resource
end
