--[[-----------------------------------------------------------------

EnergyMonitor Options defintions

This file adds the setupOptions function to the addon which gets
called during OnInitialization.

Author: lokicoyote
File Date: 2008-12-04T22:43:43Z
File Revision: 71
Project Revision: 73
Project Version: v0.10.4

--]]-----------------------------------------------------------------

local kAddon = "EnergyMonitor"
local kModule = "Options"

-- Cache often used global functions
local ipairs, pairs, setmetatable, type = _G.ipairs, _G.pairs, _G.setmetatable, _G.type
local InterfaceOptionsFrame_OpenToCategory = _G.InterfaceOptionsFrame_OpenToCategory

---------------------------------------------------------------------
-- Library constants
---------------------------------------------------------------------

local addon = LibStub("AceAddon-3.0"):GetAddon(kAddon):GetModule(kModule)
local L = LibStub("AceLocale-3.0"):GetLocale(kAddon .. "_" .. kModule, true)

---------------------------------------------------------------------
-- Local constants
---------------------------------------------------------------------

local kSlashCommand = "emon"
local kEnergyMax = 130
local kEnergyPerSecond = 10
local kSharedMediaInfo = {
	["BarEdge"] = { "border", "LSM30_Border" }
	, ["BarFont"] = { "font", "LSM30_Font" }
	, ["BarTexture"] = { "statusbar", "LSM30_Statusbar" }
	, ["SoundFile"] = { "sound", "LSM30_Sound" }
	, ["SoundResource"] = { "playsound" }
}

---------------------------------------------------------------------
-- getBaseOptions functions
---------------------------------------------------------------------

local getBaseOptions -- forward declaration
do
	local createCommandShow -- forward declaration
	do
		local function setShowWhen(info, in_enable)
			info.handler.addon:SetShowWhen(info.arg, in_enable)
		end

		local function getShowWhen(info)
			return info.handler.addon:GetShowWhen(info.arg)
		end

		function createCommandShow(in_show, in_order)
			return {
				type = "toggle"
				, order = in_order
				, name = L[ "show.when-" .. in_show ]
				, desc = L[ "show.when-" .. in_show .. "-desc" ]
				, set = setShowWhen
				, get = getShowWhen
				, arg = in_show
			}
		end
	end

	local createCommandBeat -- forward declaration
	do
		local function setBeat(info, ...)
			info.handler.addon:SetTick(info.handler.var.kReversedFire[ info.arg ])
			info.handler.addon[ "Set" .. info.arg ](info.handler.addon, ...)
		end

		local function setIsBeat(info, in_value)
			if in_value then
				setBeat(info, info.handler.addon[ "Get" .. info.arg ](info.handler.addon))
			end
		end

		local function getIsBeat(info)
			return info.handler.addon:GetTick() == info.handler.var.kReversedFire[ info.arg ]
		end

		function createCommandBeat(in_command, in_func_name, in_value)
			return {
				[ in_command .. "_gui" ] = {
					type = "group"
					, cmdHidden = true
					, name = L[ "beat-" .. in_func_name ]
					, args = {
						used = {
							type = "toggle"
							, order = 1
							, name = L[ "beat.used" ]
							, desc = L[ "beat-" .. in_func_name .. "-use" ]
							, set = setIsBeat
							, get = getIsBeat
							, arg = in_func_name
						}, value = setmetatable({
								type = "range"
								, order = 2
								, name = L[ "beat-" .. in_func_name ]
								, desc = L[ "beat-" .. in_func_name .. "-desc" ]
								, arg = in_func_name
							}, { __index = in_value })
					}
				}, [ in_command ] = setmetatable({
						type = "range"
						, guiHidden = true
						, name = L[ "beat-" .. in_func_name ]
						, desc = L[ "beat-" .. in_func_name .. "-desc" ]
						, set = setBeat
						, arg = in_func_name
					}, { __index = in_value })
			}
		end
	end

	local function setOption(info, ...)
		info.handler.addon[ "Set" .. info.arg ](info.handler.addon, ...)
	end

	local function getOption(info, ...)
		return info.handler.addon[ "Get" .. info.arg ](info.handler.addon, ...)
	end

	local hasSharedMedia -- forward declaration
	local getMediaValues -- forward declaration
	local getMediaOption -- forward declaration
	local getMediaControl -- forward declaration
	local getMediaDisabled -- forward declaration
	do
		local SharedMedia = LibStub:GetLibrary("LibSharedMedia-3.0", true)
		hasSharedMedia = SharedMedia and true

		if hasSharedMedia then
			do
				local AceGUI = LibStub:GetLibrary("AceGUI-3.0", true)
				local hasControl = setmetatable({}, {
						__index = function(t, k)
							if type(k) ~= "string" or "" == k then return nil end

							local present = AceGUI and AceGUI:Create(k) and k or nil
							rawset(t, k, present)
							return present
						end
					})

				function getMediaControl(in_option)
					local sm_info = kSharedMediaInfo[ in_option ]
					return sm_info and hasControl[ sm_info[ 2 ] ] or nil
				end
			end

			function getMediaValues(info)
				local sm_info = kSharedMediaInfo[ info.arg ]
				local sm_type = sm_info and sm_info[ 1 ]
				if not SharedMedia:IsValid(sm_type) then
					return {}
				elseif getMediaControl(info.arg) then
					return SharedMedia:HashTable(sm_type)
				else
					return SharedMedia:List(sm_type)
				end
			end

			function getMediaOption(info, ...)
				local media = getOption(info, ...)
				if type(media) == "string" and "" ~= media then
					local sm_info = kSharedMediaInfo[ info.arg ]
					local sm_type = sm_info and sm_info[ 1 ]
					if SharedMedia:IsValid(sm_type) then
						for k, v in ipairs(SharedMedia:List(sm_type)) do
							if media == SharedMedia:Fetch(sm_type, v, true) then
								if getMediaControl(info.arg) then
									return v
								else
									return k
								end
							end
						end
					end
				end
				return media
			end

			function getMediaDisabled(info)
				local sm_info = kSharedMediaInfo[ info.arg ]
				return sm_info and SharedMedia:GetGlobal(sm_info[ 1 ]) and true
			end
		else
			function getMediaControl() return nil end
			function getMediaValues() return {} end
			getMediaOption = getOption
			function getMediaDisabled() return false end
		end
	end

	-- Description: Function to return the top level Ace3 config table for the Dialog interface
	-- Expected result: Returns the top level Ace3 config table
	-- Input: None
	-- Output: Ace3 config table
	function getBaseOptions(self)
		local options_base = self.cache.options_base
		if not options_base then
			local show_when = {}
			for k, v in pairs(self.addon:Get_kShow()) do
				show_when[ k ] = createCommandShow(k, v)
			end

			local shown_while = {}
			for k, _ in pairs(self.addon:Get_kShowWhile()) do
				shown_while[ k ] = L[ "show.while-" .. k ]
			end

			options_base = {
				handler = self
				, type = "group"
				, name = L[ kAddon ]
				, set = setOption
				, get = getOption
				, args = {
					locked = {
						type = "toggle"
						, order = 1
						, name = L["locked"]
						, desc = L["locked-desc"]
						, arg = "Locked"
					}, bar = {
						type = "group"
						, order = 10
						, name = L["bar"]
						, desc = L["bar-desc"]
						, args = {
							invert = {
								type = "toggle"
								, order = 1
								, width = "half"
								, name = L["bar.invert"]
								, desc = L["bar.invert-desc"]
								, arg = "BarInverted"
							}, color = {
								type = "color"
								, order = 10
								, width = "half"
								, name = L["bar.color"]
								, desc = L["bar.color-desc"]
								, hasAlpha = true
								, arg = "BarColor"
							}, text = {
								type = "input"
								, order = 20
								, name = L["bar.text"]
								, desc = L["bar.text-desc"]
								, usage = L["bar.text-usage"]
								, arg = "BarText"
							}, font_gui = {
								type = "select"
								, hidden = not hasSharedMedia
								, cmdHidden = true
								, order = 21
								, disabled = getMediaDisabled
								, values = getMediaValues
								, get = getMediaOption
								, name = L["bar.font"]
								, arg = "BarFont"
								, dialogControl = getMediaControl("BarFont")
							}, font = {
								type = "input"
								, guiHidden = hasSharedMedia
								, order = 21
								, name = L["bar.font"]
								, arg = "BarFont"
							}, texture_gui = {
								type = "select"
								, hidden = not hasSharedMedia
								, cmdHidden = true
								, order = 30
								, disabled = getMediaDisabled
								, values = getMediaValues
								, get = getMediaOption
								, name = L["bar.texture"]
								, arg = "BarTexture"
								, dialogControl = getMediaControl("BarTexture")
							}, texture = {
								type = "input"
								, guiHidden = hasSharedMedia
								, order = 30
								, name = L["bar.texture"]
								, arg = "BarTexture"
							}, edge_gui = {
								type = "select"
								, hidden = not hasSharedMedia
								, cmdHidden = true
								, order = 35
								, disabled = getMediaDisabled
								, values = getMediaValues
								, get = getMediaOption
								, name = L["bar.edge"]
								, arg = "BarEdge"
								, dialogControl = getMediaControl("BarEdge")
							}, edge = {
								type = "input"
								, guiHidden = hasSharedMedia
								, order = 35
								, name = L["bar.edge"]
								, arg = "BarEdge"
							}, border = {
								type = "input"
								, hidden = true
								, order = 36
								, name = L["bar.border"]
								, arg = "BarBorder"
							}, width = {
								type = "range"
								, order = 40
								, name = L["bar.width"]
								, desc = L["bar.width-desc"]
								, min = 50
								, max = 300
								, step = 1
								, arg = "BarWidth"
							}, height = {
								type = "range"
								, order = 41
								, name = L["bar.height"]
								, desc = L["bar.height-desc"]
								, min = 10
								, max = 25
								, step = 1
								, arg = "BarHeight"
							}, scale = {
								type = "range"
								, order = 50
								, name = L["bar.scale"]
								, desc = L["bar.scale-desc"]
								, min = 0.25
								, max = 3.0
								, step = 0.01
								, isPercent = true
								, arg = "BarScale"
							}
						}
					}, beats = {
						type = "group"
						, order = 20
						, name = L["beats"]
						, desc = L["beats-desc"]
						, childGroups = "tab"
						, args = {}
						, plugins = {
							energy = createCommandBeat("energy", "EnergyPerTick", { min = 1, max = kEnergyMax, step = 1 })
							, percent = createCommandBeat("percent", "PercentPerTick", { min = 0.01, max = 1, step = 0.01, isPercent = true })
							, time = createCommandBeat("time", "TimePerTick", { min = 0.1, max = kEnergyMax / kEnergyPerSecond, step = 0.1 })
						}
					}, show = {
						type = "group"
						, order = 30
						, name = L["show"]
						, desc = L["show-desc"]
						, childGroups = "tab"
						, args = {
							when = {
								type = "group"
								, name = L["show.when"]
								, desc = L["show.when-desc"]
								, args = show_when
							}, while_gui = {
								type = "group"
								, cmdHidden = true
								, name = L["show.while"]
								, args = {
									["while"] = { -- Quote "while" since it's reserved
										type = "select"
										, name = L["show.while"]
										, desc = L["show.while-desc"]
										, values = shown_while
										, arg = "ShownWhile"
									}
								}
							}, ["while"] = { -- Quote "while" since it's reserved
								type = "select"
								, guiHidden = true
								, name = L["show.while"]
								, desc = L["show.while-desc"]
								, values = shown_while
								, arg = "ShownWhile"
							}
						}
					}, sound = {
						type = "group"
						, order = 40
						, name = L["sound"]
						, desc = L["sound-desc"]
						, args = {
							enabled = {
								type = "toggle"
								, name = L["sound.enabled"]
								, desc = L["sound.enabled-desc"]
								, arg = "SoundEnabled"
							}, file_gui = {
								type = "select"
								, hidden = not hasSharedMedia
								, disabled = getMediaDisabled
								, values = getMediaValues
								, get = getMediaOption
								, name = L["sound.file"]
								, desc = L["sound.file-desc"]
								, arg = "SoundFile"
								, dialogControl = getMediaControl("SoundFile")
							}, file = {
								type = "input"
								, guiHidden = hasSharedMedia
								, name = L["sound.file"]
								, desc = L["sound.file-desc"]
								, arg = "SoundFile"
							}, resource_gui = {
								type = "select"
								, hidden = not hasSharedMedia
								, cmdHidden = true
								, disabled = getMediaDisabled
								, values = getMediaValues
								, get = getMediaOption
								, name = L["sound.resource"]
								, desc = L["sound.resource-desc"]
								, arg = "SoundResource"
							}, resource = {
								type = "input"
								, guiHidden = hasSharedMedia
								, name = L["sound.resource"]
								, desc = L["sound.resource-desc"]
								, usage = L["sound.resource-usage"]
								, arg = "SoundResource"
							}
						}
					}
				}
			}
			self.cache.options_base = options_base
		end
		return options_base
	end
end

-- Description: Function to return the top level Ace3 config table for the command line
-- Expected result: Returns the top level Ace3 config table
-- Input: None
-- Output: Ace3 config table
local function getFullOptions(self)
	local options = self.cache.options
	if not options then
		options = setmetatable({ plugins = {} }, { __index = getBaseOptions(self) })
		-- Add plugins
		local AceDBOptions = LibStub:GetLibrary("AceDBOptions-3.0", true)
		if AceDBOptions then
			options.plugins[ "profile" ] = {
				profile = AceDBOptions:GetOptionsTable(self.addon.db)
			}
		end
		self.cache.options = options
	end
	return options
end

local chatCommand -- forward declaration
do
	local AceConfigCmd, HasAceConfigDialog
	if LibStub:GetLibrary("AceConfig-3.0", true) then
		AceConfigCmd = LibStub:GetLibrary("AceConfigCmd-3.0", true)
		HasAceConfigDialog = LibStub:GetLibrary("AceConfigDialog-3.0", true) and true
	end

	local function chatCommandLineOnly(self, in_input)
		AceConfigCmd:HandleCommand(kSlashCommand, kAddon .. "Full", in_input)
	end

	local function chatCommandGuiOnly(self, in_input)
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
	end

	if HasAceConfigDialog then
		if AceConfigCmd then
			function chatCommand(self, in_input)
				if in_input and in_input:trim() ~= "" then
					chatCommandLineOnly(self, in_input)
				else
					chatCommandGuiOnly(self, in_input)
				end
			end
		else
			chatCommand = chatCommandGuiOnly
		end
	else
		if AceConfigCmd then
			chatCommand = chatCommandLineOnly
		else
			function chatCommand()
				error("Unable to process slash commands")
			end
		end
	end
end

---------------------------------------------------------------------
-- Setup functions
---------------------------------------------------------------------

function addon:setupOptions()
	self.var.kReversedFire = {}
	for k, v in pairs(self.addon:Get_kFire()) do
		self.var.kReversedFire[ v ] = k
	end

	self.ChatCommand = function(_, in_input)
		chatCommand(self, in_input)
	end

	local AceConfig = LibStub:GetLibrary("AceConfig-3.0", true)
	if not L or not AceConfig then
		return
	end

	AceConfig:RegisterOptionsTable(kAddon, getBaseOptions(self))
	AceConfig:RegisterOptionsTable(kAddon .. "Full", getFullOptions(self))
	for k, v in pairs(getFullOptions(self).plugins) do
		AceConfig:RegisterOptionsTable(kAddon .. " " .. L[ k ], v[ k ])
	end

	local AceConfigDialog = LibStub:GetLibrary("AceConfigDialog-3.0", true)
	if AceConfigDialog then
		self.optionsFrame = AceConfigDialog:AddToBlizOptions(kAddon, L[ kAddon ])
		self.optionsFrameProfile = AceConfigDialog:AddToBlizOptions(kAddon .. " " .. L[ "profile" ], L[ "profile" ], L[ kAddon ])
	end
end

---------------------------------------------------------------------
-- LibSharedMedia-3.0 Handlers
---------------------------------------------------------------------

function addon:LibSharedMedia_Register()
	local AceConfigRegistry = LibStub:GetLibrary("AceConfigRegistry-3.0", true)
	if AceConfigRegistry then
		AceConfigRegistry:NotifyChange(kAddon)
		AceConfigRegistry:NotifyChange(kAddon .. "Full")
	end
end
